from collections import namedtuple
import abc
from types import FunctionType

from myo.logging import Logging
from myo.output.data import OutputEvent, OutputEntry
from myo.record import Record

from fn.recur import tco

from amino import List, Maybe, Try, curried, Just, L, _

from ribosome.record import re_field, field


class ParserBase(Logging, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def events(self, output: List[str]) -> List[OutputEvent]:
        ...


class EdgeData(Record):
    r = re_field()
    entry = field((type, FunctionType))

Step = namedtuple('Step', ['node', 'data', 'weight'])


# TODO log errors
@curried
def cons(entry, match):
    return Try(entry, text=match.string, **match.groupdict()).to_maybe


class SimpleParser(ParserBase, metaclass=abc.ABCMeta):

    @abc.abstractproperty
    def graph(self):
        ...

    def event(self, entries: List[OutputEntry]):
        return Just(OutputEvent(entries=entries))

    def edges(self, node):
        e = (Step(t, d['data'], d.get('weight', 0))
             for f, t, d in self.graph.edges(node, data=True))
        return List.wrap(e)

    def match_edge(self, node, match):
        return (
            self.edges(node)
            .sort_by(_.weight, reverse=True)
            .find_map(match)
        )

    @tco
    def _process(self, node: str, output: List[str], result: List[OutputEvent],
                 current: List[OutputEntry]) -> List[OutputEvent]:
        '''
        Parse a list of output lines. Uses a trampoline for tail call
        elimination.
        The algorithm starts at the graph node 'start'
        1. detach the first output line into *line* and *rest* and call
           *parse_line*
        2. find an edge that matches to current line
        a) if a match was found
            3. construct an *OutputEntry*
            4. set the current node to the target node of the edge
            5. add the entry to the list *current*
            6. recurse with *rest* as new *output*
        b) if no match was found
            7. construct an *OutputEvent* from *current*
            8. if the current node is 'start', set *output* to *rest*
               else, keep *output* to try again with 'start'
            9. recurse with 'start'
        10. add the last event and exit the recursion
        '''
        def add_event():
            new = current.empty.no.flat_maybe_call(L(self.event)(current))
            return result + new.to_list
        def parse_line(line: str, rest: List[str]):
            self.log.debug('parsing line: {}'.format(line))
            def match(step: Step):
                def dbg(a):
                    self.log.debug('matched edge to {}'.format(step.node))
                return (
                    Maybe(step.data.r.match(line)) %
                    dbg //
                    cons(step.data.entry)
                ) & Just(step.node)
            def cont(entry: OutputEntry, next_node: str):
                return True, (self, next_node, rest, result,
                              current.cat(entry))
            def next_event():
                new_output = (rest if node == 'start' and current.empty else
                              output)
                return True, (self, 'start', new_output, add_event(), List())
            return self.match_edge(node, match).map2(cont) | next_event
        quit = lambda: (False, add_event())
        return output.detach_head.map2(parse_line) | quit

    def events(self, output):
        return self._process(self, 'start', output, List(), List())

__all__ = ('ParserBase',)



from ribosome.machine import message, json_message

JumpCurrent = message('JumpCurrent')
JumpCursor = message('JumpCursor')
Jump = message('Jump', 'target')
SetLoc = message('SetLoc', 'index', opt_fields=(('jump', True),))
InitialError = message('InitialError')
DisplayLines = message('DisplayLines')
SetResult = json_message('SetResult', 'result')
ToggleFilters = message('ToggleFilters')
EventNext = message('EventNext')
EventPrev = message('EventPrev')
CursorToCurrent = message('CursorToCurrent')

__all__ = ('JumpCurrent', 'JumpCursor', 'Jump', 'SetLoc', 'InitialError',
           'DisplayLines', 'SetResult', 'ToggleFilters', 'EventNext',
           'EventPrev', 'CursorToCurrent')



from amino import Right, Left, Map, Maybe, __

from myo.util.ident import Ident, ident_field


def parse_int(i):
    return Right(i) if isinstance(i, int) else (
        Right(int(i)) if isinstance(i, str) and i.isdigit() else
        Left('could not parse int {}'.format(i))
    )


def parse_id(value, rex, desc):
    return (
        Maybe(rex.match(str(value)))
        .map(__.group(1))
        .map(int)
        .to_either("could not match {} id {}".format(desc, value))
        .or_else(lambda: parse_int(value)))


def amend_options(opt: Map, key: str, value: Maybe):
    return value / (lambda a: opt + (key, a)) | opt

__all__ = ('parse_int', 'parse_id', 'Ident', 'ident_field')



from functools import singledispatch  # type: ignore

from myo.command import ShellCommand, TransientCommand


@singledispatch
def unite_format(obj):
    return str(obj)

unite_format_str_command = '[{name}] {line}'


def _format_cmd(name, display_name, line):
    word = unite_format_str_command.format(name=display_name, line=line)
    return dict(word=word, name=name)


@unite_format.register(ShellCommand)
def unite_format_command(cmd):
    return _format_cmd(cmd.name, cmd.name, cmd.line)


@unite_format.register(TransientCommand)
def unite_format_transient_command(cmd):
    return _format_cmd(cmd.name, cmd.command.name, cmd.line)

__all__ = ('unite_format',)



from pathlib import Path

import neovim

from amino import List, L, _, I, Map

from ribosome import command, NvimStatePlugin, msg_command, json_msg_command
from ribosome.machine.scratch import Mapping
from ribosome.request import msg_function
import ribosome.nvim.components
from ribosome.machine.state import UpdateRecord
from ribosome.unite import mk_unite_candidates, mk_unite_action
from ribosome.unite.plugin import unite_plugin

from myo.plugins.core.main import StageI
from myo.main import Myo
from myo.logging import Logging
from myo.plugins.command.message import (AddVimCommand, Run, AddShellCommand,
                                         AddShell, ShellRun, RunTest,
                                         RunVimTest, CommandShow, RunLatest,
                                         RunLine)
from myo.plugins.tmux.message import (TmuxCreatePane, TmuxCreateSession,
                                      TmuxCreateLayout, TmuxSpawnSession,
                                      TmuxInfo, TmuxClosePane, TmuxPack,
                                      TmuxMinimize, TmuxRestore, TmuxToggle,
                                      TmuxFocus, TmuxOpenOrToggle, TmuxKill)
from myo.plugins.tmux import TmuxOpen
from myo.plugins.core.message import Parse, Resized
from myo.plugins.unite.message import UniteHistory
from myo.plugins.unite.main import UniteNames
from myo.plugins.unite.format import unite_format
from myo.output.machine import EventPrev, EventNext

unite_candidates = mk_unite_candidates(UniteNames)
unite_action = mk_unite_action(UniteNames)


@unite_plugin('myo')
class MyoNvimPlugin(NvimStatePlugin, Logging):

    def __init__(self, vim: neovim.Nvim) -> None:
        super().__init__(vim)
        self.myo = None
        self._post_initialized = False

    @property
    def name(self):
        return 'myo'

    @property
    def state(self):
        return self.myo

    @command()
    def myo_reload(self):
        self.myo_quit()
        self.myo_start()

    @command()
    def myo_quit(self):
        if self.myo is not None:
            self.myo.stop()
            self.vim.clean()
            self.myo = None

    @command(sync=True)
    def myo_start(self):
        config_path = self.vim.vars.ppath('config_path')\
            .get_or_else(Path('/dev/null'))
        plugins = self.vim.vars.pl('plugins') | self._default_plugins
        self.myo = Myo(self.vim.proxy, Path(config_path), plugins)
        self.myo.start()
        self.myo.wait_for_running()
        self.myo.send(StageI())

    @property
    def _default_plugins(self):
        return List('command', 'tmux', 'unite')

    @command()
    def myo_post_startup(self):
        self._post_initialized = True
        if self.myo is not None:
            pass
        else:
            self.log.error('myo startup failed')

    @command()
    def myo_plug(self, plug_name, cmd_name, *args):
        self.myo.plug_command(plug_name, cmd_name, args)

    @json_msg_command(AddVimCommand)
    def myo_vim_command(self):
        pass

    @json_msg_command(AddShellCommand)
    def myo_shell_command(self):
        pass

    @json_msg_command(AddShell)
    def myo_shell(self):
        pass

    @json_msg_command(TmuxCreateSession)
    def myo_tmux_create_session(self):
        pass

    @msg_command(TmuxSpawnSession)
    def myo_tmux_spawn_session(self):
        pass

    @json_msg_command(TmuxCreateLayout)
    def myo_tmux_create_layout(self):
        pass

    @json_msg_command(TmuxCreatePane)
    def myo_tmux_create_pane(self):
        pass

    @json_msg_command(TmuxOpen)
    def myo_tmux_open(self):
        pass

    @msg_command(TmuxClosePane)
    def myo_tmux_close_pane(self):
        pass

    @msg_command(TmuxPack)
    def myo_tmux_pack(self):
        pass

    @json_msg_command(TmuxMinimize)
    def myo_tmux_minimize(self):
        pass

    @json_msg_command(TmuxRestore)
    def myo_tmux_restore(self):
        pass

    @json_msg_command(TmuxToggle)
    def myo_tmux_toggle(self):
        pass

    @msg_command(TmuxOpenOrToggle)
    def myo_tmux_open_or_toggle(self):
        pass

    @msg_command(TmuxFocus)
    def myo_tmux_focus(self):
        pass

    @json_msg_command(TmuxKill)
    def myo_tmux_kill(self):
        pass

    @json_msg_command(Run)
    def myo_run(self):
        pass

    @msg_command(RunLine)
    def myo_run_line(self):
        pass

    @json_msg_command(ShellRun)
    def myo_run_in_shell(self):
        pass

    @json_msg_command(RunLatest)
    def myo_run_latest(self):
        pass

    @msg_command(TmuxInfo)
    def myo_tmux_show(self):
        pass

    @msg_command(CommandShow)
    def myo_command_show(self):
        pass

    @json_msg_command(Parse)
    def myo_parse(self):
        pass

    @msg_command(EventPrev)
    def myo_event_prev(self):
        pass

    @msg_command(EventNext)
    def myo_event_next(self):
        pass

    @json_msg_command(RunTest)
    def myo_test(self):
        pass

    @json_msg_command(RunVimTest)
    def myo_vim_test(self):
        pass

    @json_msg_command(UpdateRecord)
    def myo_update(self):
        pass

    def _eval(self, args, go):
        result = List.wrap(args).head.to_either('expression needed') // go
        return result.cata(self.log.error, I)

    @neovim.function('MyoEval', sync=True)
    def myo_eval(self, args):
        return self._eval(args, self.myo.eval_expr)

    @neovim.function('MyoTmuxEval', sync=True)
    def myo_tmux_eval(self, args):
        def mod(data, plugins):
            d = data.sub_states.get('tmux') | data
            p = plugins.get('tmux') | plugins
            return d, p
        return self._eval(args, L(self.myo.eval_expr)(_, mod))

    @neovim.autocmd('VimLeave', sync=True)
    def vim_leave(self):
        ribosome.nvim.components.shutdown = True
        self.myo_quit()

    @neovim.autocmd('VimResized', sync=False)
    def vim_resized(self):
        if self.myo is not None:
            self.myo.send(Resized())

    @msg_function(Mapping)
    def myo_mapping(self):
        pass

    @msg_command(UniteHistory)
    def myo_unite_history(self):
        pass

    @unite_candidates('history')
    def myo_unite_history_candidates(self, args):
        return self.myo.data.commands.history / unite_format

    @unite_action('run', 'name')
    def myo_unite_run(self, ident):
        return Run(ident, options=Map())

__all__ = ('MyoNvimPlugin',)



import logging
from typing import Callable

import amino.logging
import ribosome.logging
from ribosome.logging import ribosome_logger, pr

from amino.lazy import lazy


log = myo_root_logger = ribosome_logger('myo')


def myo_logger(name: str):
    return myo_root_logger.getChild(name)


class Logging(amino.logging.Logging):

    @lazy
    def _log(self) -> amino.logging.Logger:  # type: ignore
        return myo_logger(self.__class__.__name__)


def print_info(out: Callable[[str], None]):
    lname = lambda l: logging.getLevelName(l.getEffectiveLevel())
    hlname = lambda h: logging.getLevelName(h.level)
    def handler(h):
        return '{}({})'.format(h.__class__.__name__, hlname(h))
    def logger(l):
        handlers = ','.join(list(map(handler, l.handlers)))
        info = '{}: {} {}'.format(l.name, lname(l), handlers)
        out(info)
    logger(amino.logging.amino_root_logger)
    logger(ribosome.logging.ribosome_root_logger)
    logger(myo_root_logger)

__all__ = ('myo_logger', 'Logging')



from uuid import UUID
from typing import Union

from amino import Maybe, List, L, _, Empty

from ribosome.record import field, Record, uuid_field


class Key(Record):
    uuid = uuid_field()
    name = field(str)

    @property
    def _str_extra(self):
        return super()._str_extra.cat(self.name)

Ident = Union[str, UUID, Key]


def contains_pane_ident(a: Maybe):
    tpes = List(UUID, str)
    bad = lambda a: not tpes.exists(L(isinstance)(a, _))
    err = 'must be Maybe[Ident]'
    return not a.exists(bad), err


def ident_field():
    return field(Maybe, initial=Empty(), invariant=contains_pane_ident)

__all__ = ('Ident', 'contains_pane_ident', 'ident_field', 'Key')



from amino.test import later
from amino import _

from myo.ui.tmux.facade.window import pane_not_found_error
from myo.plugins.tmux.main import invalid_pane_name

from integration._support.tmux import (DefaultLayoutSpec, TmuxIntegrationSpec,
                                       ExternalTmuxIntegrationSpec)


class CreateLayoutSpec(ExternalTmuxIntegrationSpec):

    def create(self):
        self.plugin.myo_start()
        name = 'lay'
        self._create_layout(name, parent='root')
        l = self.state.sessions.head // _.windows.head // _.root.layouts.last
        (l / _.name).should.contain(name)


class CreatePaneSpec(ExternalTmuxIntegrationSpec):

    def create(self):
        self.plugin.myo_start()
        lname = 'lay'
        pname = 'pan'
        self._create_layout(lname, parent='root')
        self._create_pane(pname, parent=lname)
        p = (self.state.sessions.head // _.windows.head //
             _.root.layouts.last // _.panes.head)
        (p / _.name).should.contain(pname)


class XOpenSpec(ExternalTmuxIntegrationSpec):

    def create(self):
        self.vim.vars.set_p('tmux_no_watcher', True)
        self.plugin.myo_start()
        lname = 'lay'
        pname = 'pan'
        self._create_layout(lname, parent='root')
        self._create_pane(pname, parent=lname)
        self._open_pane(pname)
        self._pane_count(2)


class OpenPaneSpec(TmuxIntegrationSpec):

    def open(self):
        pan = 'pan'
        self.json_cmd('MyoTmuxCreateLayout lay', parent='root')
        self._create_pane(pan, parent='lay')
        self._open_pane(pan)
        self._pane_count(2)
        pan1 = 'pan1'
        self._open_pane(pan1)
        self._log_contains(invalid_pane_name.format(pan1))


class ClosePaneSpec(DefaultLayoutSpec):

    def close(self):
        pane = 'make'
        self._open_pane(pane)
        self._pane_count(2)
        self._close_pane(pane)
        self._pane_count(1)
        self._close_pane(pane)
        self._pane_count(1)
        self._log_contains(pane_not_found_error.format(pane))
        self._open_pane(pane)
        self._pane_count(2)
        self._close_pane(pane)
        self._pane_count(1)

    def auto_close(self):
        self._open_pane('make')
        self._pane_count(2)
        self.vim.doautocmd('VimLeave')
        self._pane_count(1)


class ShowSpec(TmuxIntegrationSpec):

    def output(self):
        def check():
            self._log_out.should_not.be.empty
        self.vim.cmd('MyoTmuxShow')
        later(check)


class PinSpec(TmuxIntegrationSpec):

    def open(self):
        check = lambda: self._panes.length.should.equal(3)
        def check2():
            (self._panes.find(_.height == 2) / _.active).should.contain(True)
        self.json_cmd('MyoTmuxCreateLayout test', parent='root')
        self._create_pane('pan1', parent='test', pin=True)
        self._create_pane('pan2', parent='test', focus=True, fixed_size=2)
        self._open_pane('pan2')
        later(check)
        later(check2)


class FocusSpec(TmuxIntegrationSpec):

    def open_no_focus(self):
        check = lambda i: (self._panes.lift(i) / _.active).should.contain(True)
        self._create_pane('pan1')
        self._open_pane('pan1')
        self._pane_count(2)
        later(check, 0)
        self.vim.cmd_sync('MyoTmuxFocus pan1')
        later(check, 1)

    def open_focus(self):
        check = lambda: (self._panes.lift(1) / _.active).should.contain(True)
        self._create_pane('pan1', focus=True)
        self._open_pane('pan1')
        later(check)


class OpenLayoutSpec(TmuxIntegrationSpec):

    def _pre_start(self):
        super()._pre_start()
        self.json_cmd('MyoTmuxCreateLayout lay', parent='root')
        self._create_pane('pan1', parent='lay')

    def open(self):
        self._open_pane('lay')
        self._pane_count(2)

    def toggle(self):
        self.json_cmd('MyoTmuxOpenOrToggle lay')
        self._pane_count(2)

    def minimize(self):
        self.json_cmd('MyoTmuxCreateLayout lay2', parent='root')
        self._create_pane('pan2', parent='lay2')
        self._open_pane('lay')
        self._open_pane('lay2')
        self._pane_count(3)
        self.json_cmd('MyoTmuxMinimize lay2')
        later(lambda: (self._pane_with_id(2) / _.width).should.contain(2))

    def open_twice(self):
        self._create_pane('pan', parent='lay')
        self.vim.cmd_sync('MyoTmuxOpen lay')
        self._pane_count(2)
        self.vim.cmd_sync('MyoTmuxOpen lay')
        self._wait(1)
        self._pane_count(2)


class UpdateVimWindowSpec(ExternalTmuxIntegrationSpec):

    def window_size(self):
        self.plugin.myo_start()
        self.wait_for_state()
        sz = self._window.size
        later(lambda:
              (self.state.vim_window // _.size).should.contain(sz)
              )

__all__ = ('ClosePaneSpec', 'ShowSpec')



from ribosome.machine import message, json_message

TmuxRunCommand = json_message('TmuxRunCommand', 'command')
TmuxRunShell = json_message('TmuxRunShell', 'shell')
TmuxRunCommandInShell = message('TmuxRunCommandInShell', 'shell', 'command',
                                'options')
TmuxRunLineInShell = json_message('TmuxRunLineInShell', 'shell')
TmuxRunTransient = json_message('TmuxRunTransient', 'cmd')
TmuxCreateSession = json_message('TmuxCreateSession')
TmuxSpawnSession = message('TmuxSpawnSession', 'id')
TmuxCreateLayout = json_message('TmuxCreateLayout', 'name')
TmuxCreatePane = json_message('TmuxCreatePane', 'name')
TmuxOpen = json_message('TmuxOpen', 'name')
TmuxClosePane = message('TmuxClosePane', 'name')
TmuxFindVim = message('TmuxFindVim')
TmuxLoadDefaults = message('TmuxLoadDefaults')
TmuxInfo = message('TmuxInfo')
TmuxParse = json_message('TmuxParse')
TmuxPack = message('TmuxPack')
TmuxPostOpen = json_message('TmuxPostOpen', 'ident')
TmuxPostCommand = message('TmuxPostCommand', 'command', 'pane_ident')
TmuxFixFocus = message('TmuxFixFocus', 'pane')
TmuxFocus = message('TmuxFocus', 'pane')
TmuxMinimize = json_message('TmuxMinimize', 'pane')
TmuxRestore = json_message('TmuxRestore', 'pane')
TmuxToggle = json_message('TmuxToggle', 'pane')
TmuxOpenOrToggle = json_message('TmuxOpenOrToggle', 'pane')
TmuxKill = json_message('TmuxKill', 'pane')
UpdateVimWindow = message('UpdateVimWindow')

StartWatcher = message('StartWatcher')
QuitWatcher = message('QuitWatcher')
WatchCommand = message('WatchCommand', 'command', 'pane')

__all__ = ('TmuxRunCommand', 'TmuxOpen', 'TmuxCreatePane', 'TmuxCreateSession',
           'TmuxCreateLayout', 'TmuxCreateLayout', 'TmuxSpawnSession',
           'TmuxFindVim', 'TmuxLoadDefaults', 'TmuxClosePane',
           'TmuxRunCommandInShell', 'TmuxRunLineInShell', 'TmuxRunShell',
           'StartWatcher', 'WatchCommand', 'QuitWatcher', 'TmuxParse',
           'TmuxPack', 'TmuxPostOpen', 'TmuxFixFocus', 'TmuxMinimize',
           'TmuxRestore', 'TmuxToggle', 'TmuxFocus', 'TmuxOpenOrToggle',
           'TmuxKill', 'UpdateVimWindow', 'TmuxRunTransient',
           'TmuxPostCommand')



from ribosome.machine import may_handle, Machine, handle, Nop
from ribosome.nvim import ScratchBuffer, NvimFacade
from ribosome.machine.scratch import ScratchMachine, Quit
from ribosome.machine.base import UnitTask
from ribosome.record import (field, list_field, either_field, any_field,
                             bool_field, dfield)
from ribosome.machine.transition import Fatal

from amino.task import Task
from amino import (Map, _, L, Left, __, List, Either, Just, Try, Right,
                   Boolean, Empty)
from amino.lazy import lazy

from myo.output.data import ParseResult, Location, OutputLine
from myo.state import MyoTransitions
from myo.logging import Logging
from myo.record import Record
from myo.output.reifier.base import Reifier, LiteralReifier
from myo.output.syntax.base import LineGroups
from myo.output.filter.base import DuplicatesFilter
from myo.output.message import (JumpCurrent, JumpCursor, Jump, SetLoc,
                                InitialError, DisplayLines, SetResult,
                                ToggleFilters, EventNext, EventPrev,
                                CursorToCurrent)

error_filtered_result_empty = 'filtered result is empty'


def fold_callbacks(cbs, z):
    return cbs.fold_left(Right(z))(lambda z, a: z // L(Try)(a, _))


class ResultAdapter(Record):
    buffer = any_field()
    result = field(ParseResult)
    filters = list_field()
    reifier = either_field(Reifier)
    formatters = list_field()
    syntaxes = list_field()
    initial_error = list_field()

    @lazy
    def vim(self):
        return self.buffer.root

    @property
    def _filters(self):
        return self.filters.cat(DuplicatesFilter(self.vim))

    @lazy
    def filtered(self):
        return fold_callbacks(self._filters, self.result)

    def effective_result(self, filtered: Boolean):
        return self.filtered if filtered else Right(self.result)

    def format(self, lines: List[OutputLine]):
        return fold_callbacks(self.formatters, lines)

    @property
    def _reifier(self):
        return (
            self.reifier.o(lambda: self._lang_reifier) |
            (lambda: LiteralReifier(self.buffer))
        )

    @property
    def _lang_reifier(self):
        want = self.vim.vars.pb('lang_reifier') | Boolean(True)
        def create(lang):
            return (
                Either.import_name('myo.output.reifier.{}'.format(lang),
                                   'Reifier')
            )
        return want.maybe(self.langs) // __.find_map(create) / __(self.buffer)

    @lazy
    def filtered_lines(self):
        return self.filtered // L(Try)(self._reifier, _) // self.format

    @lazy
    def unfiltered_lines(self):
        return Try(self._reifier, self.result) // self.format

    def lines(self, filtered: Boolean):
        return (self.effective_result(filtered) // L(Try)(self._reifier, _) //
                self.format)

    @property
    def langs(self):
        return self.result.langs

    def __repr__(self):
        return repr(self.result)

    def locations(self, filtered: Boolean):
        return self.effective_result(filtered) / _.locations


class OMState(Record):
    result = field(ResultAdapter)
    locations = list_field(Location)
    lines = list_field(OutputLine)
    filter_state = bool_field()
    current_loc_index = dfield(0)


def _jump_first(locs):
    return Empty() if locs.empty else Just(0)


def _jump_last(locs):
    return Empty() if locs.empty else Just(locs.length - 1)


def _jump_default(langs, lang_jumps):
    return langs.find_map(lang_jumps.get) | (lambda: _jump_first)


class OutputMachineTransitions(MyoTransitions):

    @may_handle(SetResult)
    def set_result(self):
        result = self._adapter(self.msg.result)
        return (
            self.with_sub(self.state.set(result=result, filter_state=True)),
            DisplayLines()
        )

    @handle(DisplayLines)
    def display_lines(self):
        return self.set_content.map2(
            lambda a, b: (a, b.replace(Just(InitialError()))))

    @handle(InitialError)
    def initial_error(self):
        jump = self.vim.vars.pb('jump_to_error') | True
        locs = self.locations
        return (
            self.result.initial_error.find_map(__(locs)) /
            L(SetLoc)(_, jump)
        )

    @may_handle(Jump)
    def jump(self):
        open_file = self.open_file(self.msg.target.file_path)
        l, c = self.msg.target.coords
        set_line = Task.delay(lambda: self.vim.window.set_cursor(l, c))
        post = Task.delay(self.vim.cmd, 'normal! zvzz')
        return UnitTask(open_file + set_line + post)

    @handle(JumpCurrent)
    def jump_current(self):
        return self.current_loc / Jump

    @handle(JumpCursor)
    def jump_cursor(self):
        return self.location_for_cursor // self.index_for_location / SetLoc

    @may_handle(ToggleFilters)
    def toggle_filters(self):
        new_state = self.state.filter_state.no
        return (
            self.with_sub(self.state.set(filter_state=new_state)),
            DisplayLines()
        )

    @handle(EventNext)
    def event_next(self):
        return self.cycle_loc(_ < self.state.locations.length, _ + 1)

    @handle(EventPrev)
    def event_prev(self):
        return self.cycle_loc(_ > 0, _ - 1)

    @may_handle(SetLoc)
    def set_loc(self):
        return (
            self.with_sub(self.state.set(current_loc_index=self.msg.index)),
            CursorToCurrent(),
            (JumpCurrent() if self.msg.jump else Nop())
        )

    @handle(CursorToCurrent)
    def cursor_to_current(self):
        return (
            self.line_for_current_loc /
            L(Task.delay)(self.vim.window.set_cursor, _ + 1) /
            UnitTask
        )

    def _adapter(self, result):
        filters = self._callbacks('output_filters')
        reifier = (self._callback('output_reifier')
                   .to_either('no reifier specified'))
        formatters = self._callbacks('output_formatters')
        syntaxes = self._callbacks('output_syntaxes', self.buffer)
        initial_error = (self._callbacks('initial_error',
                                         special=self.special_jumps)
                         .cat(_jump_default(result.langs, self.lang_jumps)))
        return ResultAdapter(buffer=self.buffer, result=result,
                             filters=filters, reifier=reifier,
                             formatters=formatters, syntaxes=syntaxes,
                             initial_error=initial_error)

    @property
    def scratch(self):
        return self.machine.scratch

    @property
    def buffer(self):
        return self.scratch.buffer

    @property
    def window(self):
        return self.scratch.window

    @property
    def result(self):
        return self.state.result

    @property
    def special_jumps(self):
        return Map(first=_jump_first, last=_jump_last)

    @property
    def lang_jumps(self):
        return Map(python=_jump_last)

    @property
    def lines(self):
        return self.state.lines

    @property
    def locations(self):
        return self.state.locations

    @property
    def set_content(self):
        min_size = self.vim.vars.pi('output_window_min_size') | 3
        max_size = self.vim.vars.pi('output_window_max_size') | 30
        def run(locations, lines):
            text = lines / _.formatted
            size = min(max_size, max(min_size, text.length))
            return (
                self._with_sub(self.data,
                               self.state.set(locations=locations,
                                              lines=lines)),
                Task.delay(self.window.cmd, 'resize {}'.format(size)) +
                Task.delay(self.buffer.set_modifiable, True) +
                Task.delay(self.buffer.set_content, text) +
                Task.delay(self.buffer.set_modifiable, False) +
                self.run_syntax(lines)
            )
        def check(locations, lines):
            return (
                Left(Fatal(error_filtered_result_empty))
                if lines.empty else
                Right(run(locations, lines))
            )
        f = self.state.filter_state
        return (self.result.locations(f) &
                self.result.lines(f)).flat_map2(check)

    @property
    def syntax(self):
        lang = self.lang_syntax.to_list
        return self.result.syntaxes.cons(LineGroups(self.vim)) + lang

    @property
    def lang_syntax(self):
        def create(lang):
            return Either.import_name('myo.output.syntax.{}'.format(lang),
                                      'Syntax')
        return self.result.langs.find_map(create) / __(self.vim)

    def run_syntax(self, lines):
        return (self.syntax / (lambda a: a(lines))).sequence(Task)

    def target_for_line(self, line):
        return self.lines.lift(line) / _.target

    def line_for_target(self, target):
        return self.lines.index_where(_.target == target)

    def line_for_location(self, index):
        def same_location(a, b):
            return isinstance(a, Location) and a.same_location(b.target)
        return (
            self.locations.lift(index) //
            (lambda a: self.lines.index_where(L(same_location)(a, _))))

    def index_for_location(self, loc):
        return self.locations.index_where(__.same_location(loc))

    @property
    def location_for_cursor(self):
        return (
            self.vim.window.line /
            (_ - 1) //
            self.target_for_line
        ).filter_type(Location).to_either('not a location')

    @property
    def current_loc(self):
        return self.locations.lift(self.state.current_loc_index)

    @property
    def line_for_current_loc(self):
        return self.line_for_location(self.state.current_loc_index)

    def cycle_loc(self, cond, update):
        cur = self.state.current_loc_index
        return Boolean(cond(cur)).m(update(cur)) / SetLoc

    def open_file(self, path):
        def split():
            # TODO
            pass
        if not path.is_file():
            return Task.now(Left('not a file: {}'.format(path)))
        else:
            win = (
                self.vim.windows.find(__.buffer.modifiable.contains(True))
                .o(split)
                .task('could not get a window') /
                __.focus()
            )
            edit = Task.delay(self.vim.edit, path) / __.run_async()
            return win + edit


class OutputMachine(ScratchMachine, Logging):
    Transitions = OutputMachineTransitions

    def __init__(self, vim: NvimFacade, scratch: ScratchBuffer,
                 result: ParseResult, parent: Machine, options: Map) -> None:
        super().__init__(vim, scratch, parent=parent, title='output')
        self._result = result
        self._options = options

    @property
    def options(self):
        return self._options

    @property
    def prefix(self):
        return 'Myo'

    @property
    def mappings(self):
        return Map({
            '%cr%': JumpCursor,
            'f': ToggleFilters,
            'q': Quit,
        })

    def new_state(self):
        return OMState(result=ResultAdapter(buffer=self.scratch.buffer,
                                            result=ParseResult()))

__all__ = ('OutputMachine',)



from amino import List, __, Just, _, L
from amino.test import later
from amino.lazy import lazy

from ribosome.test.unite import unite
from ribosome.record import encode_json

from myo.command import ShellCommand, TransientCommand
from myo.plugins.unite.format import unite_format, unite_format_str_command

from integration._support.command import CmdSpec

indent = ' {}'.format


class _UniteSpecBase(CmdSpec):

    @property
    def _plugins(self):
        return super()._plugins + List('myo.plugins.unite',
                                       'integration._support.plugins.dummy')


class UniteLoadHistorySpec(_UniteSpecBase):

    @property
    def _cmd(self):
        return 'test1'

    @lazy
    def _history(self):
        cmd = ShellCommand(name='other', line='ls')
        return List(ShellCommand(name=self._cmd, line='tee'),
                    TransientCommand(command=cmd, line='tail'))

    def _set_vars(self):
        super()._set_vars()
        self.vim.vars.set('Myo_history',
                          encode_json(self._history).get_or_raise)

    @property
    def _last(self):
        return (lambda: self.vim.vars.pd('last_command') // __.get('name'))

    @unite
    def history(self):
        self._create_command(self._cmd, 'ls')
        self.vim.cmd_sync('MyoUniteHistory')
        target = self._history / unite_format / _['word'] / indent
        later(lambda: self.vim.buffer.content.should.equal(target))


class UniteHistorySpec(_UniteSpecBase):

    @unite
    def history(self):
        name1 = 'com1'
        name2 = 'com2'
        cmd = 'ls'
        self._create_command(name1, cmd)
        self._create_command(name2, cmd)
        self._run_command(name1)
        self._run_command(name2)
        form = L(unite_format_str_command).format() >> indent
        later(lambda: self._last().should.contain(name2))
        self.vim.cmd_sync('MyoUniteHistory')
        target = List(form(name=name2, line=cmd), form(name=name1, line=cmd))
        later(lambda: self.vim.buffer.content.should.equal(target))
        self.vim.window.set_cursor(2)
        self.vim.feedkeys('\r')
        later(lambda: self._last().should.contain(name2))
        self.vim.cmd_sync('MyoUniteHistory')
        later(lambda: self.vim.buffer.content.should.equal(target.reversed))

_test_line = 'echo \'testing\''


def _test_ctor():
    return Just(_test_line)


class UniteTestSpec(_UniteSpecBase):

    @unite
    def unite_test_run(self):
        self.json_cmd_sync('MyoTest',
                           ctor='py:integration.unite_spec._test_ctor')
        later(lambda: (self._last() | '').should.contain('test_line'))
        self.vim.vars.set_p('last_command', {})
        self.vim.cmd_sync('MyoUniteHistory')
        target = unite_format_str_command.format(name='<test>',
                                                 line=_test_line)
        self._buffer_content(List(indent(target)))
        self.vim.feedkeys('\r')
        later(lambda: (self._last() | '').should.contain('test_line'))

__all__ = ('UniteLoadHistorySpec', 'UniteHistorySpec')



import amino
from amino import __, F, List, Left, Task, Map

from ribosome.machine import may_handle, Error, RunTask, handle
from ribosome.machine.base import io
from ribosome.machine.transition import Fatal

from myo.state import MyoComponent, MyoTransitions
from myo.plugins.core.dispatch import VimDispatcher
from myo.plugins.core.message import StageI, Initialized, ParseOutput
from myo.output import VimCompiler, CustomOutputHandler, Parsing
from myo.output.main import OutputHandler

error_no_output_events = 'no events in parse result'


class CoreTransitions(MyoTransitions):

    @may_handle(StageI)
    def stage_1(self):
        return Initialized().pub.at(1)

    @may_handle(Initialized)
    def initialized(self):
        started = io(__.vars.set_p('started', True))
        return (self.data.set(initialized=True) + VimDispatcher(self.vim),
                started)

    @may_handle(Error)
    def error(self):
        m = self.msg.message
        handler = (F(self.log.caught_exception, 'transition') if
                   amino.development and isinstance(m, Exception) else
                   self.log.error)
        handler(m)

    @handle(ParseOutput)
    def parse_output(self):
        opt = self.msg.options
        def handle(parser):
            def display(result):
                return (
                    Task.now(Left(Fatal(error_no_output_events)))
                    if result.events.empty else
                    parser.display(result, opt)
                )
            return RunTask(parser.parse(self.msg.output, self.msg.path) //
                           display)
        return self._error_handler(self.msg.command) / handle

    def _error_handler(self, cmd):
        langs = self.msg.options.get('langs') / List.wrap | cmd.langs
        parser_spec = self.msg.options.get('parser').o(cmd.parser)
        def check(p):
            return (p if isinstance(p, OutputHandler) else
                    CustomOutputHandler(self.vim, p))
        return (
            self._inst_callbacks(parser_spec, None, self._special_parsers) /
            check
        ).o(self._langs_parsing(langs)).lmap(Fatal)

    @property
    def _special_parsers(self):
        return Map(compiler=VimCompiler)

    def _langs_parsing(self, langs):
        return langs.empty.no.either('command has no langs',
                                     Parsing(self.vim, langs))


class Plugin(MyoComponent):
    Transitions = CoreTransitions

    def new_state(self):
        pass

__all__ = ('Plugin',)
